/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	UNICNT:					*/
/*						*/
/************************************************/
#include        <sys/mount.h>
#include        <sys/syscall.h>
#include        <sys/time.h>
#include        <sys/wait.h>
#include	<time.h>
#include	<signal.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<unistd.h>

#include	"lowtyp.h"
#include	"dbgmem.h"
#include	"subapl.h"
#include	"unicnt.h"

/*container filesystem				*/
#define	ROOTFS	"rootfs"	/*container root*/
#define	OLDROOT	".oldroot"	/*for pivot_root*/
#define	CLONPID	"first.pid"	/*cont process 1*/
#define	MTAB	"/etc/mtab"	/*mtab file	*/
/*

*/
/************************************************/
/*						*/
/*	Defining pivot_root system call		*/
/*						*/
/************************************************/
static int pivot_root(const char * new_root,const char * put_old)

{
#ifndef __NR_pivot_root
#pragma message ("the pivot_root syscall is not available within \"sys/syscall.h\"")
#pragma message ("-> The pivot_root system call will generate an alert")
(void) apl_alert(0,"%s need the pivot_root system call!!!.",appname);
errno=ENOSYS;
return -1;
#else
return(syscall(__NR_pivot_root, new_root, put_old));
#endif
}
/*

*/
/************************************************/
/*						*/
/*	Returning the container path.		*/
/*	contpath array is dynamically assigned	*/
/*	need to be freed when not need anymore.	*/
/*						*/
/************************************************/
static char *getcontpath(char *contname)

{
char *contpath;
char *vzpath;

contpath=(char *)0;
vzpath=apl_appdir(d_vzgot);
(void) asprintf(&contpath,"%s/%s",vzpath,contname);
(void) free(vzpath);
return contpath;
}
/*

*/
/************************************************/
/*						*/
/*	Check is the container is here with	*/
/*	all needed parts.			*/
/*						*/
/************************************************/
int cnt_iscontgood(char *contname)

{
int isgood;
char *contpath;
int phase;
int proceed;

isgood=true;
phase=0;
proceed=true;
contpath=getcontpath(contname);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is container existing?*/
      if (contpath!=(char *)0) {/*always.	*/
	char *rootpath;	

	(void) asprintf(&rootpath,"%s/%s",contpath,ROOTFS);
        if (apl_isdir(rootpath)<0) {
          (void) apl_alert(0,"%s, rootfs directory missing within container \"cont\" directory",
			     appname,contname);
	  isgood=false;
	  phase=999;	/*trouble trouble	*/
	  }
	(void) free(rootpath);
	}
      break;
    case 1	:	/*check/make oldroot dir*/
      if (contpath!=(char *)0) {/*always.	*/
	char *oldpath;

	(void) asprintf(&oldpath,"%s/%s",contpath,OLDROOT);
        if (apl_isdir(oldpath)<0) {
	  if (mkdir(oldpath,0700)<0) {
            (void) apl_alert(0,"%s, unable to make %s directory for container <%s> (error=%s)",
				appname,OLDROOT,contname,strerror(errno));
	    isgood=false;
	    phase=999;	/*trouble trouble	*/
	    }
	  }
	(void) free(oldpath);
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(contpath);
return isgood;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to set an initial mtab	*/
/*	contents WITHIN the clone rootfs	*/
/*						*/
/************************************************/
static int setmtab(char *contname)

{
int done;
FILE *fichier;
int phase;
int proceed;

done=false;
fichier=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*removing mtab file	*/
      if (unlink(MTAB)<0) {
	switch (errno)	{
	  case ENOENT	:/* file not existing:-}*/
	    break;
	  default	:
            (void) apl_alert(0,"%s, Unable to remove %s from container <%s> rootfs (error=<%s>)",
		     	        appname,MTAB,contname,strerror(errno));
	    phase=999;	/*trouble trouble	*/
	    break;
	  }
	}
      break;
    case 1	:	/*creating new MTAB file*/
      if ((fichier=fopen(MTAB,"w"))==(FILE *)0) {
        (void) apl_alert(0,"%s, Unable to create %s within container <%s> rootfs (error=<%s>)",
		     	    appname,MTAB,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*inserting contents	*/
      if (fprintf(fichier,"%s / roofs rw 0 0\n",appname)<0) {
	(void) apl_alert(0,"%s, Unable to write %s file in container <%s> (error=<%s>)",
			    appname,MTAB,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 3	:	/*everything fine	*/
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to pivot the container root	*/
/*	file system.				*/
/*						*/
/************************************************/
int cnt_pivot(int foreground,char *contname)

{
int done;
char *contpath;
int phase;
int proceed;

done=false;
contpath=getcontpath(contname);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*binding container file*/
      if (mount(contpath,contpath,0,MS_BIND,0)<0) {
        (void) apl_alert(0,"%s, unable to bind container <%s> directory (error=%s)",
				appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*moving to rootfs	*/
      (void) cnt_setstdio(foreground,contname,"container.log");
      if (chdir(contpath)<0) {
        (void) apl_alert(0,"%s, unable to reach container <%s> directory (error=%s)",
				appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*moving to rootfs	*/
      if (pivot_root(".",OLDROOT)<0) {
        (void) apl_alert(0,"%s, unable to pivot container <%s> root (error=%s)",
				appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*umounting previous fs	*/
      if (umount2(OLDROOT,MNT_DETACH)<0) {
        (void) apl_alert(0,"%s, unable to unmount container <%s> oldroot (error=%s)",
				appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/* chrooting to rootfs	*/
      if (chdir(ROOTFS)<0) {
        (void) apl_alert(0,"%s, unable to reach container <%s> rootfs (error=%s)",
				appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 5	:	/* everything fine	*/
      if (chroot(".")<0) {
        (void) apl_alert(0,"%s, unable to rootfs container <%s> (error=%s)",
				appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 6	:	/* everything fine	*/
      (void) setmtab(contname);	/*initiat mtab	*/
      (void) fprintf(stdout,"%s container chroot done\n",apl_ascsystime(time((time_t *)0)));
      (void) fflush(stdout);
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to set a file with the clone	*/
/*	ip number.				*/
/*						*/
/************************************************/
int cnt_setclonepid(char *contname,pid_t cpid)

{
int done;
FILE *fichier;
char *filename;
char *contpath;
int phase;
int proceed;

done=false;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the file	*/
      if ((fichier=fopen(filename,"w"))==(FILE *)0) {
	(void) apl_alert(0,"%s, Unable to open pidfile <%s> (error=<%s>)",
			    appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*writing  PID in it	*/
      if (fprintf(fichier,"%05d\n",cpid)<0) {
	(void) apl_alert(0,"%s, Unable to write pidfile in container <%s> (error=<%s>)",
			    appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*everything fine	*/
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get the clone pid stored	*/
/*	within the clonpid file			*/
/*						*/
/************************************************/
pid_t cnt_getclonepid(char *contname)

{
pid_t clonepid;
FILE *fichier;
char *filename;
char *contpath;

clonepid=(pid_t)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
if ((fichier=fopen(filename,"r"))!=(FILE *)0) {
  char strloc[80];

  (void) fgets(strloc,sizeof(strloc)-1,fichier);
  (void) sscanf(strloc,"%d",&clonepid);
  (void) fclose(fichier);
  }
(void) free(filename);
(void) free(contpath);
return clonepid;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to remove the file with the	*/
/*	clone pid.				*/
/*						*/
/************************************************/
int cnt_rmclonepid(char *contname)

{
int done;
char *filename;
char *contpath;

done=true;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
if (unlink(filename)<0) {
  (void) apl_alert(0,"%s, Unable to remove pidfile in container <%s> (error=<%s>)",
		     appname,contname,strerror(errno));
  done=false;
  }
(void) free(filename);
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to initialize the HOST side	*/
/*						*/
/************************************************/
int cnt_initscript(char *scriptname,const char *fmt,...)

{
#define	BSIZE	2048

int status;
va_list args;
FILE *canal;
char *varlibpath;
char *parmlst;
char *cmd;
int phase;
int proceed;

status=0;
va_start(args,fmt);
canal=(FILE *)0;
varlibpath=apl_appdir(d_varlib);
(void) vasprintf(&parmlst,fmt,args);
(void) asprintf(&cmd,"%s/shell/%s %s",varlibpath,scriptname,parmlst);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening pip channel	*/
      if ((canal=popen(cmd,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s, Unable to pipe cmd <%s> (error=<%s>)",
			    appname,cmd,strerror(errno));
	status=1;
	phase=999;	/*trouble trouble	*/	
	}
      break;
    case 1	:	/*do we have feed back	*/
      if (canal!=(FILE *)0) {	/*always	*/
	char buffer[BSIZE];

	while (fgets(buffer,BSIZE,canal)!=(char *)0) {
	  (void) apl_cleanstring(buffer);
	  if ((debug>0)&&(verbose==true))
	    (void) apl_alert(0,"%s, pipe say: <%s>",appname,buffer);
	  else {
	    (void) fprintf(stdout,"%s\n",buffer);
	    (void) fflush(stdout);
	    }
	  }
	}
      break;
    case 2	:	/*closing pip		*/
      status=pclose(canal);/*get pipe status	*/
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(cmd);
(void) free(parmlst);
(void) free(varlibpath);
va_end(args);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to extract the container	*/
/*	architecture available in the 'arch' 	*/
/*	file.					*/
/*						*/
/************************************************/
char *cnt_getarch(char *contname)

{
static char *availarch[]={
	"i386","i686","x86_64",(char *)0
	};

char *arch;
FILE *fichier;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[200];

arch=(char *)0;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"arch");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*let open the file	*/
      if ((fichier=fopen(filename,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			   appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*reading first line	*/
      if (apl_getstr(fichier,buffer,sizeof(buffer),'#')==(char *)0) {
	(void) apl_alert(0,"%s Unable to read file <%s> (file empty?)",
			   appname,filename);
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*scanning line		*/
      (void) apl_cleanstring(buffer);
      if (strlen(buffer)>0) { 	/*always?	*/
	int i;

        for (i=0;availarch[i]!=(char *)0;i++) {
	  if (strcmp(availarch[i],buffer)==0) {
	    arch=strdup(availarch[i]);
	    proceed=false;	/*found arch	*/
	    break;
	    }
	  }
	}
      break;
    case 3	:	/*scanning line		*/
      (void) apl_alert(0,"%s file <%s>, arch <%s> unknown, setting default",
			   appname,filename,buffer);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
if (arch==(char *)0) {
  arch=strdup(availarch[0]);
  }
return arch;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to extract the container	*/
/*	distribution type available in the  	*/
/*	'dist' file.				*/
/*						*/
/************************************************/
char *cnt_getdist(char *contname)

{
char *dist;
FILE *fichier;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[200];

dist=(char *)0;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"dist");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*let open the file	*/
      if ((fichier=fopen(filename,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			   appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*reading first line	*/
      if (apl_getstr(fichier,buffer,sizeof(buffer),'#')==(char *)0) {
	(void) apl_alert(0,"%s Unable to read file <%s> (file empty?)",
			   appname,filename);
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*scanning line		*/
      (void) apl_cleanstring(buffer);
      dist=strdup(buffer);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
if (dist==(char *)0) {
  dist=strdup("unknown");
  }
return dist;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to assign a new STDOUT/STDERR	*/
/*	and close STDIN. Done ONLY if the	*/
/*	process is in background mode.		*/
/*						*/
/************************************************/
void cnt_setstdio(int foreground,char *contname,char *outname)

{
#define	OPIO	"unicnt.c:cnt_setstdio,"

if (foreground==0) {
  int newstdout;
  int newstderr;
  int phase;
  int proceed;
  char *contpath;
  char *filename;

  newstdout=-1;
  newstderr=-1;
  phase=0;
  proceed=true;
  contpath=getcontpath(contname);
  (void) asprintf(&filename,"%s/%s.stdout",contpath,outname);
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*open the new stdout	*/
	(void) unlink(filename);
	if ((newstdout=open(filename,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP))<0) {
	  (void) apl_alert(0,"%s, Unable to open <%s> (error=<%s>)",
			      OPIO,filename,strerror(errno));
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case 1	:	/*open the new stderr	*/
	(void) free(filename);
        (void) asprintf(&filename,"%s/%s.stderr",contpath,outname);
	(void) unlink(filename);
	if ((newstderr=open(filename,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP))<0) {
	  (void) apl_alert(0,"%s, Unable to open <%s> (error=<%s>)",
			      OPIO,filename,strerror(errno));
	  (void) close(newstdout);
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case 2	:	/*duplicating stdout	*/
	if (dup2(newstdout,1)<0) {
	  (void) apl_alert(0,"%s, Unable to dup2 stdout (error=<%s>)",
			      OPIO,strerror(errno));
	  (void) close(newstderr);
	  (void) close(newstdout);
	  phase=999;
	  }
	break;
      case 3	:	/*duplicating stderr	*/
	(void) close(newstdout);
	if (dup2(newstderr,2)<0) {
	  (void) apl_alert(0,"%s, Unable to dup2 stderr (error=<%s>)",
			      OPIO,strerror(errno));
	  (void) close(newstderr);
	  phase=999;
	  }
	break;
      case 4	:	/*setting new std	*/
	if ((stdout=fdopen(1,"w"))==(FILE *)0) {
	  (void) apl_alert(0,"%s, Unable to fdopen stdout (error=<%s>)",
			      OPIO,strerror(errno));
	  }
	if ((stderr=fdopen(2,"w"))==(FILE *)0) {
	  (void) apl_alert(0,"%s, Unable to fdopen stderr (error=<%s>)",
			      OPIO,strerror(errno));
	  }
	(void) fclose(stdin);
	break;
      case 5	:	/*write time stamp	*/
	(void) fprintf(stdout,"stdout start: %s\n",apl_ascsysdatetime(time((time_t *)0)));
	(void) fprintf(stderr,"stderr start: %s\n",apl_ascsysdatetime(time((time_t *)0)));
	(void) fflush(stdout);
	(void) fflush(stderr);
	break;
      default	:	/*SAFE Guard		*/
	proceed=false;
        break;
      }
    phase++;
    }
  (void) free(filename);
  (void) free(contpath);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for entry on the	*/
/*	container console and forward it	*/
/*	on the Master_container console file.	*/
/*	A Management expiration is done		*/
/*	according time.				*/
/*						*/
/************************************************/
int cnt_mstconsole(char *contname,pid_t cntpid)

{
#define	OPCO	"unicnt.c:cnt_console,"

#define	RLX	5	/*number of sec		*/
int status;
int nbr;
int cconsole;
int mconsole;
struct timeval relax;
fd_set reading;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[1000];

status=0;
nbr=0;
cconsole=0;
mconsole=0;
relax.tv_sec=RLX;	/*sec relax time	*/
relax.tv_usec=0;
FD_ZERO(&reading);
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"console");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the M-Console*/
      if ((mconsole=open(filename,O_CREAT|O_WRONLY|O_TRUNC,S_IRUSR|S_IWUSR))<0) {
	(void) apl_alert(0,"%s, Unable to open mconsole <%s> (error=<%s>)",
			      OPCO,filename,strerror(errno));
	phase=999;
	status=-1;
	}
      break;
    case 1	:	/*opening the C-Console*/
      (void) free(filename);
      (void) asprintf(&filename,"%s/rootfs/dev/%s",contpath,"console");
      if ((cconsole=open(filename,O_RDWR|O_ASYNC))<0) {
	(void) apl_alert(0,"%s, Unable to open cconsole <%s> (error=<%s>)",
			      OPCO,filename,strerror(errno));
  	(void) close(mconsole);
	phase=999;
	status=-1;
        }
      break;
    case 2	:	/*waiting event		*/
      relax.tv_sec=RLX;	/*sec relax time	*/
      relax.tv_usec=0;
      FD_ZERO(&reading);
      FD_SET(cconsole,&reading);
      switch ((nbr=select(cconsole+1,&reading,(fd_set *)0,(fd_set *)0,&relax))) {
        case	-1    :
	  switch (errno) {
	    case EINTR:	/*received a signal, lets see...*/
	      break;	/*getting out fast		*/
	    default   :
	      (void) apl_alert(0,"%s, unexpected select event (error=<%s>)",
			          OPCO,strerror(errno));
	      break;
	    }
          break;
        case      0   :	/*timeout child lost ?	*/
	  switch (waitpid(cntpid,0,WNOHANG)) {
	    case -1   :	/*exiting 		*/
	      (void) apl_alert(0,"Unexpect error on waitpid (error=<%s>)",strerror(errno));
	      status=-1;
	      break;
	    case 0    :	/*process still up!	*/
	      phase--;	/*lets wait for console	*/
	      break;
	    default   :	/*process exited!	*/
			/*going to next phase	*/
	      break;
	    }
          break;
	default	      :	/*data ready in cconsole*/
	  if ((nbr=read(cconsole,buffer,sizeof(buffer)))>0) {
	    (void) write(mconsole,buffer,nbr);
	    }		/*forwarded to mconsole	*/
	  phase--;	/*stay put on phase	*/
          break;
	}
      break;
    case 3	:	/*closing console	*/
      (void) close(mconsole);
      (void) close(cconsole);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
return status;
}
